#!/usr/bin/env python3

# Internal packages
from dataclasses import dataclass
from typing import Final


@dataclass
class ProgramVersion:
    """
    Contains the program version information.
    """

    major: Final[int] = 1
    minor: Final[int] = 0
    patch: Final[int] = 0
    beta: Final[bool] = False

    def __repr__(self):
        return f"{self.major}.{self.minor}.{self.patch}{'b' if self.beta else ''}"


# md2anki general information
MD2ANKI_NAME: Final = "md2anki"
MD2ANKI_AUTHOR: Final = "AnonymerNiklasistanonym"
MD2ANKI_DESCRIPTION: Final = "Convert Markdown formatted documents to anki decks"
MD2ANKI_URL: Final = "https://github.com/AnonymerNiklasistanonym/Md2Anki"
MD2ANKI_URL_GIT: Final = f"{MD2ANKI_URL}.git"
MD2ANKI_URL_BUG_TRACKER: Final = f"{MD2ANKI_URL}/issues"
MD2ANKI_VERSION: Final = ProgramVersion(major=3, minor=0, patch=18, beta=True)

# md2anki Markdown preprocessor information
MD2ANKI_MD_PP_ANKI_DECK_HEADING_SUBDECK_PREFIX: Final = "Subdeck: "
MD2ANKI_MD_PP_ANKI_NOTE_QUESTION_ANSWER_SEPARATOR: Final = "---"
MD2ANKI_MD_PP_EVALUATE_CODE_LANGUAGE_PREFIX: Final = "="
MD2ANKI_MD_PP_COMMENT_INSERT_FILE_PREFIX = f"{MD2ANKI_NAME.upper()}_INSERT_FILE="
MD2ANKI_MD_PP_MD2ANKI_TAG_PREFIX = "{=:"
MD2ANKI_MD_PP_MD2ANKI_TAG_SUFFIX = ":=}"

# md2anki evaluate code information
MD2ANKI_EVALUATE_CODE_ENV_NAME_ANKI_HTML_BOOL = f"{MD2ANKI_NAME.upper()}_ANKI_HTML"
MD2ANKI_EVALUATE_CODE_ENV_NAME_PANDOC_PDF_BOOL = f"{MD2ANKI_NAME.upper()}_PANDOC_PDF"



#!/usr/bin/env python3

# Internal packages
import sys as _sys

# Local modules
from md2anki.cli import parse_cli_args
from md2anki.main import main


def entry_point():
    cli_args = parse_cli_args(_sys.argv[1:])
    exit_code = main(cli_args)
    exit(exit_code)



#!/usr/bin/env python3

# Internal packages
import os
from typing import Dict, Final, List, Tuple

# Local modules
from md2anki.info.general import MD2ANKI_NAME


# Types
EvaluateCodeLanguageId = str
EvaluateCodeCommand = str
EvaluateCodeCommandArgument = str
EvaluateCodeInfo = Dict[
    EvaluateCodeLanguageId,
    List[Tuple[EvaluateCodeCommand, List[EvaluateCodeCommandArgument]]],
]

# Evaluate code information
EVALUATE_CODE_MD2ANKI_PLACEHOLDER_CODE_STRING: Final = f"{MD2ANKI_NAME.upper()}_CODE"
EVALUATE_CODE_MD2ANKI_PREFIX_CODE_FILE_NAME: Final = (
    f"{MD2ANKI_NAME.upper()}_CODE_FILE="
)
EVALUATE_CODE_DEFAULT_COMMANDS: Final[EvaluateCodeInfo] = {
    "py": [("python", ["-c", EVALUATE_CODE_MD2ANKI_PLACEHOLDER_CODE_STRING])],
    "js": [("node", ["-e", EVALUATE_CODE_MD2ANKI_PLACEHOLDER_CODE_STRING])],
    "ts": [
        (
            "ts-node" + (".cmd" if os.name == "nt" else ""),
            [f"{EVALUATE_CODE_MD2ANKI_PREFIX_CODE_FILE_NAME}code.ts"],
        )
    ],
    "pl": [
        (
            "swipl",
            [
                "-O",
                "-s",
                f"{EVALUATE_CODE_MD2ANKI_PREFIX_CODE_FILE_NAME}code.pl",
                "-g",
                "true",
                "-t",
                "halt.",
            ],
        )
    ],
    "latex": [
        (
            "latexmk",
            [
                "-shell-escape",
                "-pdf",
                f"{EVALUATE_CODE_MD2ANKI_PREFIX_CODE_FILE_NAME}code.tex",
            ],
        ),
        (
            "inkscape",
            ["--export-filename=code.svg", "code.pdf"],
        ),
    ],
    "cpp": [
        (
            "clang++",
            [
                "-Wall",
                "-std=c++20",
                f"{EVALUATE_CODE_MD2ANKI_PREFIX_CODE_FILE_NAME}main.cpp",
                "-o",
                "main.exe",
            ],
        ),
        ("main.exe", []),
    ],
    "c": [
        (
            "clang",
            [
                "-std=c17",
                f"{EVALUATE_CODE_MD2ANKI_PREFIX_CODE_FILE_NAME}main.c",
                "-o",
                "main.exe",
            ],
        ),
        ("main.exe", []),
    ],
}
"""The commands that should per default be used to evaluate code of certain languages."""



#!/usr/bin/env python3

# Internal packages
import random
from uuid import uuid4


def create_unique_id() -> str:
    return str(uuid4())


def create_unique_id_int(length: int = 32) -> int:
    return random.getrandbits(length)



#!/usr/bin/env python3

# Internal packages
import unittest
import sys
from os import name
from pathlib import Path
from typing import Set, List, Tuple, Callable, Optional, Final
from urllib.parse import urlparse, ParseResult

# Append the module path for md2anki
sys.path.append(str(Path(__file__).parent.parent.joinpath("src")))

# Local modules
from md2anki.md_util import (
    md_get_used_files,
    md_get_used_md2anki_tags,
    md_update_local_filepaths,
    md_update_code_parts,
    md_update_images,
    md_update_math_sections,
)


class TestMdGetUsedFiles(unittest.TestCase):
    def setUp(self):
        self.md_content_list: List[str] = list()
        self.results: List[Set[Path | ParseResult]] = list()
        self.expected: List[Set[Path | ParseResult]] = list()

        test_data: List[Tuple[str, Set[Path | ParseResult]]] = [
            ("", set()),
            ("![](path1)", {Path("path1")}),
            ("hi\n![](path1)", {Path("path1")}),
            ("hi\n![](path1)\n![](path2)", {Path("path1"), Path("path2")}),
            ("![](path1) ![](path2)", {Path("path1"), Path("path2")}),
            (
                "![](path1) ![](path2) ![](https://www.google.com/image.png)",
                {
                    Path("path1"),
                    Path("path2"),
                    urlparse("https://www.google.com/image.png"),
                },
            ),
        ]

        for test_input, test_expected in test_data:
            self.md_content_list.append(test_input)
            self.results.append(md_get_used_files(test_input))
            self.expected.append(test_expected)

    def test_used_files_set_same(self):
        for md_content, result, expected in zip(
            self.md_content_list, self.results, self.expected
        ):
            with self.subTest(md_content=md_content):
                self.assertSetEqual(
                    result, expected, f"Check if used files {result=}=={expected=}"
                )


class TestMdUpdateLocalFilepaths(unittest.TestCase):
    def setUp(self):
        self.md_content_list: List[str] = list()
        self.results: List[str] = list()
        self.expected: List[str] = list()

        new_directory: Final = Path("new_dir")
        absolute_path: Final = Path(
            "C:\\Users\\user\\Documents\\Md2Anki\\abs_graph.svg"
            if name == "nt"
            else "/mnt/c/Users/user/Documents/Md2Anki/abs_graph.svg"
        )

        test_data: List[Tuple[str, Callable[[Path], str]]] = [
            ("", lambda x: ""),
            ("![](path1)", lambda x: f"![]({x.joinpath('path1')})"),
            (
                "![](path1) ![](path2) ![](https://www.google.com/image.png)",
                lambda x: f"![]({x.joinpath('path1')}) ![]({x.joinpath('path2')}) ![](https://www.google.com/image.png)",
            ),
            (
                f"Question with matplotlib graph\n\n![]({absolute_path})\n",
                lambda x: f"Question with matplotlib graph\n\n![]({x.joinpath('abs_graph.svg')})\n",
            ),
        ]

        for test_input, test_expected in test_data:
            self.md_content_list.append(test_input)
            self.results.append(md_update_local_filepaths(test_input, new_directory))
            self.expected.append(test_expected(new_directory))

    def test_updated_filepaths_same(self):
        for md_content, result, expected in zip(
            self.md_content_list, self.results, self.expected
        ):
            with self.subTest(md_content=md_content):
                self.assertEqual(
                    result,
                    expected,
                    f"Check if updated filepaths {result=}=={expected=}",
                )


class TestMdUpdateCodeParts(unittest.TestCase):
    def setUp(self):
        self.md_content_list: List[str] = list()
        self.results: List[str] = list()
        self.expected: List[str] = list()

        test_data: List[Tuple[str, str]] = [
            ("", ""),
            ("`text`", "`ctextd`"),
            ("a `text`{.python}", "a `ctextd`{a.pythonb}"),
            (
                "Format inline code `print('inline')`{.python} is also `supported`{.txt}",
                "Format inline code `cprint('inline')d`{a.pythonb} is also `csupportedd`{a.txtb}",
            ),
            (
                "abc\n```python\nprint('hi')\n```",
                "abc\n```python\nprint('hi')\n```",
            ),
            (
                "abc\n\n```python\nprint('hi')\n```",
                "abc\n\n```apythonb\ncprint('hi')\nd\n```",
            ),
            (
                "code block indented:\n\n" "    ```python\n    print('hi')\n    ```",
                "code block indented:\n\n"
                "    ```apythonb\nc    print('hi')\nd\n    e```",
            ),
        ]

        def update_code_part(
            code: str,
            code_block: bool,
            language: Optional[str],
            code_block_indent: Optional[str],
        ):
            if language is not None:
                language = f"a{language}b"
            code = f"c{code}d"
            indent = (
                f"{code_block_indent}e"
                if code_block_indent is not None and len(code_block_indent) > 0
                else ""
            )
            if code_block:
                return f"\n{code_block_indent if len(indent) > 0 else ''}```{language}\n{code}\n{indent}```"
            else:
                return f"`{code}`" + (
                    ("{" + language + "}") if language is not None else ""
                )

        for test_input, test_expected in test_data:
            self.md_content_list.append(test_input)
            self.results.append(md_update_code_parts(test_input, update_code_part))
            self.expected.append(test_expected)

    def test_updated_code_parts_same(self):
        for md_content, result, expected in zip(
            self.md_content_list, self.results, self.expected
        ):
            with self.subTest(md_content=md_content):
                self.assertEqual(
                    result,
                    expected,
                    f"Check if updated code parts {result=}=={expected=}",
                )


class TestMdUpdateImages(unittest.TestCase):
    def setUp(self):
        self.md_content_list: List[str] = list()
        self.results: List[str] = list()
        self.expected: List[str] = list()

        test_data: List[Tuple[str, str]] = [
            (
                "![alt text](source path){ width=100px, height=200px }",
                "![aalt textb](csource pathd){ width=e100pxf, height=g200pxh }",
            ),
            ("![alt text](source path){}", "![aalt textb](csource pathd)"),
            ("![alt text](source path)", "![aalt textb](csource pathd)"),
            (
                "![alt](./source.png){ height=20px }",
                "![aaltb](c./source.pngd){ height=g20pxh }",
            ),
            (
                "![alt text](source path) ![alt text2](source path2)",
                "![aalt textb](csource pathd) ![aalt text2b](csource path2d)",
            ),
        ]

        def update_image(
            file_path: str,
            file_description: str,
            width: Optional[str],
            height: Optional[str],
        ):
            style: List[str] = list()
            style_str = ""
            if width is not None:
                style.append(f"width=e{width}f")
            if height is not None:
                style.append(f"height=g{height}h")
            if len(style) > 0:
                style_str = "{ " + ", ".join(style) + " }"
            return f"![a{file_description}b](c{file_path}d){style_str}"

        for test_input, test_expected in test_data:
            self.md_content_list.append(test_input)
            self.results.append(
                md_update_images(
                    test_input,
                    update_image,
                )
            )
            self.expected.append(test_expected)

    def test_updated_images_same(self):
        for md_content, result, expected in zip(
            self.md_content_list, self.results, self.expected
        ):
            with self.subTest(md_content=md_content):
                self.assertEqual(
                    result,
                    expected,
                    f"Check if updated code parts {result=}=={expected=}",
                )


class TestMdGetUsedMd2ankiTags(unittest.TestCase):
    def setUp(self):
        self.md_content_list: List[str] = list()
        self.results: List[Set[str]] = list()
        self.expected: List[Set[str]] = list()

        test_data: List[Tuple[str, Set[str]]] = [
            ("", set()),
            ("`{=:tag list string:=}`", {"tag_list_string"}),
            ("`{=:tag_list_string:=}`", {"tag_list_string"}),
            ("`{=:tag1,tag2,tag3:=}`", {"tag1", "tag2", "tag3"}),
        ]

        for test_input, test_expected in test_data:
            self.md_content_list.append(test_input)
            self.results.append(md_get_used_md2anki_tags(test_input))
            self.expected.append(test_expected)

    def test_used_md2anki_tags_set_same(self):
        for md_content, result, expected in zip(
            self.md_content_list, self.results, self.expected
        ):
            with self.subTest(md_content=md_content):
                self.assertSetEqual(
                    result,
                    expected,
                    f"Check if used md2anki tags {result=}=={expected=}",
                )


class TestMdUpdateMathSections(unittest.TestCase):
    def setUp(self):
        self.md_content_list: List[str] = list()
        self.results: List[str] = list()
        self.expected: List[str] = list()

        test_data: List[Tuple[str, str]] = [
            ("", ""),
            ("abc", "abc"),
            ("$ab$", "$aabb$"),
            ("$$cd$$", "$$ccdd$$"),
        ]

        def update_math_section(math_section: str, block: bool):
            if block:
                return f"$$c{math_section}d$$"
            else:
                return f"$a{math_section}b$"

        for test_input, test_expected in test_data:
            self.md_content_list.append(test_input)
            self.results.append(
                md_update_math_sections(
                    test_input,
                    update_math_section,
                )
            )
            self.expected.append(test_expected)

    def test_updated_math_sections_same(self):
        for md_content, result, expected in zip(
            self.md_content_list, self.results, self.expected
        ):
            with self.subTest(md_content=md_content):
                self.assertEqual(
                    result,
                    expected,
                    f"Check if updated math section {result=}=={expected=}",
                )



#!/usr/bin/env python3

# Future import to support Python 3.9
from __future__ import annotations

# Internal packages
import argparse
import json
import logging
import sys
import tempfile
from dataclasses import dataclass, field
from operator import attrgetter
from pathlib import Path
from pygments import __version__ as pygments_version
from pygments.lexers import get_all_lexers
from typing import Optional, List, Dict, Final, Tuple, TypeVar, Callable

# Local modules
from md2anki.info.evaluate_code import (
    EVALUATE_CODE_DEFAULT_COMMANDS,
    EVALUATE_CODE_MD2ANKI_PLACEHOLDER_CODE_STRING,
)
from md2anki.info.general import (
    MD2ANKI_VERSION,
    MD2ANKI_NAME,
    MD2ANKI_MD_PP_EVALUATE_CODE_LANGUAGE_PREFIX,
)
from md2anki.md_to_pdf import PANDOC_ARGS_PDF
from md2anki.note_models import AnkiCardModelId


class SortedArgumentDefaultsHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def add_arguments(self, actions):
        actions = sorted(actions, key=attrgetter("option_strings"))
        super(argparse.ArgumentDefaultsHelpFormatter, self).add_arguments(actions)


class MdInputFileNotFoundException(Exception):
    """Raised when a markdown input file is not found"""

    md_input_file_path: Final[Path]

    def __init__(self, md_input_file_path: Path):
        absolute_path: Final = md_input_file_path.absolute()
        super().__init__(
            f"Markdown input file was not found ({md_input_file_path!r}, {absolute_path=})"
        )
        self.md_input_file_path = md_input_file_path


@dataclass
class Md2AnkiArgs:
    """
    Contains all information of the md2anki command line arguments.
    """

    additional_file_dirs: List[Path] = field(default_factory=lambda: list())
    """Additional file directories for external file references (i.e. images)."""
    anki_card_model: AnkiCardModelId = AnkiCardModelId.DEFAULT
    """Anki card model."""
    md_input_file_paths: List[Path] = field(default_factory=lambda: list())
    """The markdown input file paths."""
    md_output_file_paths: Optional[List[Path]] = None
    """The optional output file paths of the updated markdown files (i.e. with added IDs)."""
    md_output_dir_path: Optional[Path] = None
    """The optional output dir path of the updated markdown files (i.e. with added IDs)."""
    anki_output_file_path: Optional[Path] = None
    """The output file path of the anki deck output file."""
    backup_output_dir_path: Optional[Path] = None
    """The output dir path of the backup of the anki deck."""
    pdf_output_file_path: Optional[Path] = None
    """The output file path of the anki deck."""
    md_heading_depth: int = 1
    """The default heading depth for the anki deck heading (increases for each subdeck/question by 1)."""
    custom_program: Dict[str, List[str]] = field(default_factory=lambda: dict())
    """Custom programs for languages used for code evaluation."""
    custom_program_args: Dict[str, List[List[str]]] = field(
        default_factory=lambda: dict()
    )
    """Custom program args for languages used for code evaluation."""
    evaluate_code_cache_dir_path: Optional[Path] = None
    """The optional cache dir path of the evaluated code."""

    evaluate_code = False
    """Evaluate code."""
    evaluate_code_ignore_cache = False
    """Ignore cache of already evaluated code."""
    evaluate_code_delete_cache = False
    """Delete cache of already evaluated code."""
    keep_temp_files = False
    """Remove temporary files."""
    error: Optional[MdInputFileNotFoundException] = None
    """Error message in case there was an error parsing CLI args."""


T = TypeVar("T")
U = TypeVar("U")


def convert_list_to_dict_merged(
    list_object: List[Tuple[str, T]],
    value_func: Callable[[T], U],
) -> Dict[str, List[U]]:
    dict_object: Dict[str, List[U]] = dict()
    for key, value in list_object:
        dict_object.setdefault(key, []).append(value_func(value))
    return dict_object


DEFAULT_CUSTOM_PROGRAMS: Final = [
    (key, program)
    for key, values in EVALUATE_CODE_DEFAULT_COMMANDS.items()
    for program, _ in values
] + [(key, program) for key, values in PANDOC_ARGS_PDF.items() for program, _ in values]
DEFAULT_CUSTOM_PROGRAM_ARGS: Final = [
    (key, json.dumps(program_args))
    for key, values in EVALUATE_CODE_DEFAULT_COMMANDS.items()
    for _, program_args in values
] + [
    (key, json.dumps(program_args))
    for key, values in PANDOC_ARGS_PDF.items()
    for _, program_args in values
]

log = logging.getLogger(__name__)


def str_to_str(a: str) -> str:
    return a


def json_str_to_str_list(possible_json_string: str) -> List[str]:
    try:
        possible_string_list: List[str] = json.loads(possible_json_string)
        if not all(isinstance(elem, str) for elem in possible_string_list):
            raise RuntimeError(
                f"Supplied json string was not a string list ({possible_json_string=},{possible_string_list=})"
            )
        return possible_string_list
    except ValueError as err:
        raise RuntimeError(
            f"Supplied json string {possible_json_string!r} was not a valid JSON string ({err})"
        )


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Create an anki deck file (.apkg) from one or more Markdown documents. "
        "If no custom output path is given the file name of the document (+ .apkg) is used.",
        formatter_class=SortedArgumentDefaultsHelpFormatter,
        prog=MD2ANKI_NAME,
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"%(prog)s {MD2ANKI_VERSION}"
    )
    parser.add_argument(
        "--lexers",
        action="version",
        version=f"Pygments {pygments_version} supports lexers for the following languages: "
        f" {', '.join([x[0] + ' (' + ', '.join(x[1]) + ')' for x in get_all_lexers()])}. "
        "For more information check https://pygments.org/docs/lexers.",
        help="print a list of all supported lexer languages and their aliases",
    )

    parser.add_argument(
        "-d",
        "--debug",
        dest="log_level",
        const="DEBUG",
        default="INFO",
        action="store",
        nargs="?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="custom log level to the console",
    )

    parser.add_argument(
        "-e",
        "--evaluate-code",
        action="store_true",
        help="evaluate markdown inline code/code blocks with the language prefix "
        f"{MD2ANKI_MD_PP_EVALUATE_CODE_LANGUAGE_PREFIX!r} "
        "i.e. '`print(1+1)`{=python}' or '```{=python} [newline] print(1+1) [newline] ```'",
    )
    parser.add_argument(
        "--evaluate-code-ignore-cache",
        action="store_true",
        help="ignore the cached files from previous code evaluations",
    )
    parser.add_argument(
        "--evaluate-code-delete-cache",
        action="store_true",
        help="delete all cached files from previous code evaluations",
    )
    parser.add_argument(
        "-k",
        "--keep-temp-files",
        action="store_true",
        help="keep temporary files",
    )
    parser.add_argument(
        "-anki-model",
        metavar="MODEL_ID",
        type=AnkiCardModelId,
        choices=list(AnkiCardModelId),
        default=AnkiCardModelId.DEFAULT,
        help=f"custom anki card model (%(choices)s)",
    )
    parser.add_argument(
        "-o-anki",
        metavar="APKG_FILE",
        type=Path,
        help="custom anki deck (.apkg) output file path [if not given: md input file "
        "name + .apkg]",
    )
    parser.add_argument(
        "-o-md",
        metavar="MD_FILE",
        type=Path,
        help="custom updated (and merged if multiple input files) Markdown (.md) output file path for all input files",
    )
    parser.add_argument(
        "-o-md-dir",
        metavar="MD_DIR",
        type=Path,
        help="custom output directory for all updated Markdown (.md) input files",
    )
    parser.add_argument(
        "-o-backup-dir",
        metavar="BACKUP_DIR",
        type=Path,
        help="create a backup of the anki deck (i.e. merges input files and copies external files) in a directory",
    )
    parser.add_argument(
        "-o-pdf",
        metavar="PDF_FILE",
        type=Path,
        help="create a PDF (.pdf) file of the anki deck (i.e. merges input files and removes IDs)",
    )
    parser.add_argument(
        "-evaluate-code-cache-dir",
        metavar="CACHE_DIR",
        type=Path,
        help="use a custom cache dir for code evaluations",
    )
    parser.add_argument(
        "-log-file",
        metavar="LOG_FILE",
        type=Path,
        help="log all messages to a text file (.log)",
    )
    parser.add_argument(
        "-file-dir",
        metavar="FILE_DIR",
        type=Path,
        action="extend",
        nargs="*",
        default=[],
        help="add directories that should be checked for referenced files (like relative path images)",
    )

    def check_heading_depth(x: str) -> int:
        x_num = int(x)
        if x_num < 1:
            raise argparse.ArgumentTypeError("Minimum heading depth is 1")
        return x_num

    parser.add_argument(
        "-md-heading-depth",
        metavar="HEADING_DEPTH",
        type=check_heading_depth,
        default=1,
        help="use a custom Markdown heading depth (>=1)",
    )
    parser.add_argument(
        "-custom-program",
        metavar=("language", "program"),
        type=str,
        action="append",
        nargs=2,
        default=DEFAULT_CUSTOM_PROGRAMS,
        help='use custom program for code evaluation [i.e. "py" "python3.11"]',
    )
    parser.add_argument(
        "-custom-program-args",
        metavar=("language", "program-args"),
        type=str,
        action="append",
        nargs=2,
        default=DEFAULT_CUSTOM_PROGRAM_ARGS,
        help='use custom program args for code evaluation  [i.e. "py" "[\\"-c\\",\\"'
        f"{EVALUATE_CODE_MD2ANKI_PLACEHOLDER_CODE_STRING}"
        '\\"]"]',
    )
    parser.add_argument(
        "md_input_files",
        metavar="MD_INPUT_FILE",
        type=Path,
        nargs="+",
        help="Markdown (.md) input file that contains anki deck notes",
    )
    return parser


class FormatterCleanInfo(logging.Formatter):
    def format(self, record):
        if record.levelno == logging.INFO:
            self._style._fmt = "%(message)s"
        else:
            self._style._fmt = "%(asctime)s %(levelname)s %(name)s: %(message)s"
        return super().format(record)


class Formatter(logging.Formatter):
    def format(self, record):
        self._style._fmt = "%(asctime)s %(levelname)s %(name)s: %(message)s"
        return super().format(record)


def parse_cli_args(cli_args: List[str]) -> Md2AnkiArgs:
    """
    Parse the supplied CLI arguments.

    @return: Object that contains all parsed information.
    """
    parser = get_argument_parser()
    args = parser.parse_args(cli_args)
    parsed_args = Md2AnkiArgs()

    handlers: List[logging.FileHandler | logging.StreamHandler] = []

    # Create a logging handler to stdout only for info
    hdlr_info = logging.StreamHandler(sys.stdout)
    hdlr_info.setFormatter(FormatterCleanInfo())
    hdlr_info.setLevel(getattr(logging, args.log_level))
    hdlr_info.addFilter(lambda record: record.levelno < logging.WARNING)
    handlers.append(hdlr_info)

    # Create a logging handler to stderr for warnings and errors
    hdlr_warn_err = logging.StreamHandler(sys.stderr)
    hdlr_warn_err.setFormatter(Formatter())
    hdlr_warn_err.setLevel(logging.WARNING)
    hdlr_warn_err.addFilter(lambda record: record.levelno >= logging.WARNING)
    handlers.append(hdlr_warn_err)

    if args.log_file is not None:
        # Create a logging handler to a file for everything
        hdlr_file = logging.FileHandler(args.log_file, mode="w")
        hdlr_file.setFormatter(Formatter())
        hdlr_file.setLevel(logging.DEBUG)
        handlers.append(hdlr_file)

    # This is necessary to fix the file handler
    logging.root.handlers = []

    # Setup global logging
    logging.basicConfig(
        level=logging.DEBUG,
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=handlers,
    )

    log.debug(f"{cli_args=}, {args=}")

    parsed_args.additional_file_dirs = args.file_dir
    parsed_args.anki_card_model = args.anki_model
    parsed_args.md_heading_depth = args.md_heading_depth
    parsed_args.md_input_file_paths = args.md_input_files
    parsed_args.backup_output_dir_path = args.o_backup_dir
    parsed_args.pdf_output_file_path = args.o_pdf
    parsed_args.evaluate_code = args.evaluate_code
    parsed_args.evaluate_code_ignore_cache = args.evaluate_code_ignore_cache
    parsed_args.evaluate_code_delete_cache = args.evaluate_code_delete_cache
    parsed_args.evaluate_code_cache_dir_path = args.evaluate_code_cache_dir

    if parsed_args.evaluate_code_cache_dir_path is None:
        parsed_args.evaluate_code_cache_dir_path = Path(tempfile.gettempdir()).joinpath(
            f"{MD2ANKI_NAME}_evaluate_code_cache"
        )

    if args.keep_temp_files is not None:
        parsed_args.keep_temp_files = args.keep_temp_files

    temp_custom_program = convert_list_to_dict_merged(
        args.custom_program[: len(DEFAULT_CUSTOM_PROGRAMS)], str_to_str
    )
    if args.custom_program is not None:
        parsed_args.custom_program = convert_list_to_dict_merged(
            args.custom_program[len(DEFAULT_CUSTOM_PROGRAMS) :], str_to_str
        )
    parsed_args.custom_program = {
        **temp_custom_program,
        **parsed_args.custom_program,
    }

    temp_custom_program_args = convert_list_to_dict_merged(
        args.custom_program_args[: len(DEFAULT_CUSTOM_PROGRAM_ARGS)],
        json_str_to_str_list,
    )
    if args.custom_program_args is not None:
        parsed_args.custom_program_args = convert_list_to_dict_merged(
            args.custom_program_args[len(DEFAULT_CUSTOM_PROGRAM_ARGS) :],
            json_str_to_str_list,
        )
    parsed_args.custom_program_args = {
        **temp_custom_program_args,
        **parsed_args.custom_program_args,
    }

    # If an input file is not found throw error
    for md_input_file_path in parsed_args.md_input_file_paths:
        if not md_input_file_path.is_file():
            parsed_args.error = MdInputFileNotFoundException(md_input_file_path)
            return parsed_args

    # If an additional file directory is not found throw error
    for additional_file_dir in parsed_args.additional_file_dirs:
        if not additional_file_dir.is_dir():
            log.warning(
                f"Ignore {additional_file_dir=!r} because it's not a directory!"
            )
            parsed_args.additional_file_dirs.remove(additional_file_dir)

    if args.o_anki is not None:
        # Only automatically create an anki output file path if it was not none
        parsed_args.anki_output_file_path = (
            args.o_anki if f"{args.o_anki.name}".lower() != "none" else None
        )
    else:
        parsed_args.anki_output_file_path = parsed_args.md_input_file_paths[
            0
        ].parent.joinpath(f"{parsed_args.md_input_file_paths[0].stem}.apkg")

    if args.o_md is not None:
        parsed_args.md_output_file_paths = (
            [args.o_md] if f"{args.o_md.name}".lower() != "none" else None
        )
    elif args.o_md_dir:
        parsed_args.md_output_dir_path = args.o_md_dir
    else:
        parsed_args.md_output_file_paths = parsed_args.md_input_file_paths.copy()

    return parsed_args



#!/usr/bin/env python3

# Internal packages
import copy
import logging
import re
from enum import Enum
from pathlib import Path
from typing import Optional, TextIO, List, Tuple, Final

# Local modules
from md2anki.anki_deck import AnkiDeck, AnkiNote
from md2anki.create_id import create_unique_id_int, create_unique_id
from md2anki.info.anki import ANKI_SUBDECK_SEPARATOR
from md2anki.info.general import (
    MD2ANKI_MD_PP_ANKI_DECK_HEADING_SUBDECK_PREFIX,
    MD2ANKI_MD_PP_ANKI_NOTE_QUESTION_ANSWER_SEPARATOR,
)

REGEX_MD_ANKI_DECK_HEADING: Final = re.compile(r"^(#+)\s+(.+?)(?:\s+\((\d+)\))?\s*$")
"""
Group 1: anki deck heading depth
Group 2: anki deck heading text
Group 3: anki deck id (optional)
"""
REGEX_MD_ANKI_NOTE_QUESTION_HEADING: Final = re.compile(
    r"^(##+)\s+(.+?)(?:\s+\(([^()\s]+?)\))?\s*$"
)
"""
Group 1: anki note question heading depth
Group 2: anki note question heading text
Group 3: anki note id (optional)
"""

log = logging.getLogger(__name__)


class NoAnkiDeckFoundException(Exception):
    """Raised when no anki deck is found"""

    pass


class RootAnkiDeckWasSubdeckException(Exception):
    """Raised when the root anki deck is a subdeck"""

    pass


class AnkiNoteUnexpectedDepth(Exception):
    """Raised when an anki note is found at an unexpected depth"""

    pass


class AnkiSubdeckUnexpectedDepth(Exception):
    """Raised when an anki subdeck is found at an unexpected depth"""

    pass


class ParseStateMarkdownDocument(str, Enum):
    """
    Parse states of the Markdown document parser.
    """

    LOOKING_FOR_ANKI_DECK_HEADING = ("LOOKING_FOR_ANKI_DECK_HEADING",)
    """Nothing was yet detected, anki deck heading is expected."""
    INSIDE_ANKI_DECK = ("INSIDE_ANKI_DECK",)
    """Anki deck heading was read and an anki note question/sub deck heading or anki deck description is expected."""
    LOOKING_FOR_ANKI_SUBDECK_DESCRIPTION = ("LOOKING_FOR_ANKI_SUBDECK_DESCRIPTION",)
    """Anki sub deck heading was read and an anki note question/sub deck heading or anki subdeck description is
    expected."""
    ANKI_NOTE_QUESTION = ("ANKI_NOTE_QUESTION",)
    """An anki note question heading was read and a question answer seperator or answer is expected."""
    ANKI_NOTE_ANSWER = ("ANKI_NOTE_ANSWER",)
    """An anki note question answer seperator was read and an answer is expected."""


def parse_possible_anki_deck_heading(
    md_file_line: str, parent_deck_name: Optional[str] = None
) -> Optional[Tuple[AnkiDeck, int, bool]]:
    """
    Try to parse an anki (sub) deck heading from a Markdown file.

    @param md_file_line: The Markdown line.
    @param parent_deck_name: The current parent deck name in case this is a subdeck heading.
    @return: Anki (sub) deck with heading depth and subdeck indicator if match otherwise None.
    """
    regex_match: Final = REGEX_MD_ANKI_DECK_HEADING.match(md_file_line)

    if regex_match is not None:
        anki_deck_name = regex_match.group(2)
        anki_deck_guid = (
            int(regex_match.group(3))
            if regex_match.group(3) is not None
            else create_unique_id_int()
        )
        anki_deck_is_subdeck = anki_deck_name.startswith(
            MD2ANKI_MD_PP_ANKI_DECK_HEADING_SUBDECK_PREFIX
        )
        if anki_deck_is_subdeck:
            if parent_deck_name is None:
                log.warning(
                    f"Found {anki_deck_name=} heading that is a subdeck but no parent deck name"
                )
            anki_deck_name = anki_deck_name[
                len(MD2ANKI_MD_PP_ANKI_DECK_HEADING_SUBDECK_PREFIX) :
            ]
        heading_depth = len(regex_match.group(1))
        # sanity check (sub)deck titles for Anki subdeck separator
        if ANKI_SUBDECK_SEPARATOR in anki_deck_name and (
            heading_depth == 1 or anki_deck_is_subdeck
        ):
            log.warning(
                f"The anki deck heading {anki_deck_name!r} contains the symbol {ANKI_SUBDECK_SEPARATOR!r} "
                f"used to declare a subdeck which is replaced in order to avoid errors, please use a separate "
                f"heading structure using {MD2ANKI_MD_PP_ANKI_DECK_HEADING_SUBDECK_PREFIX!r} instead"
            )
            anki_deck_name = anki_deck_name.replace(
                rf"{ANKI_SUBDECK_SEPARATOR}", ":\u200B:"
            )
        if parent_deck_name is not None:
            anki_deck_name = (
                f"{parent_deck_name}{ANKI_SUBDECK_SEPARATOR}{anki_deck_name}"
            )
        tmp_anki_deck = AnkiDeck(
            name=anki_deck_name,
            guid=anki_deck_guid,
        )
        return (
            tmp_anki_deck,
            heading_depth,
            anki_deck_is_subdeck,
        )
    return None


def parse_possible_anki_note_heading(
    md_file_line: str,
) -> Optional[Tuple[AnkiNote, int]]:
    regex_match: Final = REGEX_MD_ANKI_NOTE_QUESTION_HEADING.match(md_file_line)

    if regex_match is not None:
        return AnkiNote(
            question=regex_match.group(2),
            guid=regex_match.group(3)
            if regex_match.group(3) is not None
            else create_unique_id(),
        ), len(regex_match.group(1))
    return None


def parse_md_content_to_anki_deck_list(
    md_file: TextIO, md_file_dirs: List[Path], initial_heading_depth: int = 1
) -> List[AnkiDeck]:
    """
    Parse a Markdown file to an anki deck.

    @param md_file: The Markdown file content.
    @param md_file_dirs: The Markdown file directories of that file.
    @param initial_heading_depth: The initial Markdown heading depth.
    @return: Anki deck object list (in case sub decks are found).
    """
    anki_deck_order_index: int = 0
    """A counter to preserve the order of read anki deck headings"""
    empty_lines: int = 0
    """The amount of empty lines read"""
    parse_state: ParseStateMarkdownDocument = (
        ParseStateMarkdownDocument.LOOKING_FOR_ANKI_DECK_HEADING
    )
    """Parse state to know what information is currently being searched for"""
    anki_decks: Final[List[AnkiDeck]] = []
    """Final list of anki decks"""
    anki_decks_stack: Final[List[AnkiDeck]] = []
    """The current anki decks stack"""

    for line in md_file:
        log.debug(f"{line=!r} ({parse_state=},{empty_lines=},{anki_decks_stack=})")
        line_stripped = line.strip()

        if len(line_stripped) == 0:
            # 1. Skip empty lines but keep track of them
            empty_lines += 1
            log.debug(f"=> Found {empty_lines=}")
        elif parse_state == ParseStateMarkdownDocument.LOOKING_FOR_ANKI_DECK_HEADING:
            # 2. Look for a root anki deck heading
            possible_anki_deck = parse_possible_anki_deck_heading(line)
            if possible_anki_deck is not None:
                root_anki_deck, heading_depth, is_subdeck = possible_anki_deck
                log.debug(f"=> Found {root_anki_deck=}")
                if heading_depth != initial_heading_depth:
                    # Ignore root headings that are not in the specified initial depth
                    log.warning(
                        f"   -> Ignore {root_anki_deck=} because ({heading_depth=}!={initial_heading_depth=})"
                    )
                elif is_subdeck:
                    # Raise error if root heading is a subdeck
                    raise RootAnkiDeckWasSubdeckException(
                        f"The root anki deck heading can't be a subdeck ({line!r})"
                    )
                else:
                    # Add root anki deck to the stack
                    root_anki_deck.md_document_order = anki_deck_order_index
                    anki_deck_order_index += 1
                    anki_decks_stack.append(root_anki_deck)
                    # Change parse state to indicate that a root anki deck was found
                    parse_state = ParseStateMarkdownDocument.INSIDE_ANKI_DECK
        elif (
            parse_state == ParseStateMarkdownDocument.INSIDE_ANKI_DECK
            or parse_state == ParseStateMarkdownDocument.ANKI_NOTE_QUESTION
            or parse_state == ParseStateMarkdownDocument.ANKI_NOTE_ANSWER
        ):
            # 3. Searching for an anki subdeck heading or anki note question heading
            # - INSIDE_ANKI_DECK: or an existing anki deck description
            # - ANKI_NOTE_QUESTION: or an existing anki note answer / question (if separator is found)
            # - ANKI_NOTE_ANSWER: or an existing anki note answer
            possible_anki_sub_deck = parse_possible_anki_deck_heading(
                line, parent_deck_name=anki_decks_stack[-1].name
            )
            possible_anki_note = parse_possible_anki_note_heading(line)
            if possible_anki_sub_deck is not None and possible_anki_sub_deck[2]:
                # A subdeck was detected
                (
                    anki_subdeck,
                    heading_depth,
                    is_subdeck,
                ) = possible_anki_sub_deck
                log.debug(f">> Detected {anki_subdeck=}")
                # If there is a change in heading depth pop subdecks from the stack that won't be referenced again
                if len(anki_decks_stack) > (heading_depth - initial_heading_depth):
                    deck_was_moved = False
                    # Store the old anki deck name before updating it
                    old_anki_parentdeck_name = anki_decks_stack[-1].name
                    # While the anki deck stack has more decks than the heading depth of the found subdeck indicated
                    while len(anki_decks_stack) > (
                        heading_depth - initial_heading_depth
                    ):
                        anki_decks.append(anki_decks_stack.pop())
                        log.debug(
                            f"<= [anki decks] Append {anki_decks[-1]!r} "
                            f"[{len(anki_decks_stack)=}>{heading_depth - initial_heading_depth=}]",
                        )
                        deck_was_moved = True
                    # Refresh sub deck name in case that the previous parent deck was removed from the stack
                    if deck_was_moved:
                        anki_subdeck.name = anki_subdeck.name.replace(
                            old_anki_parentdeck_name, anki_decks_stack[-1].name, 1
                        )
                        log.debug(
                            f">> Update {anki_subdeck.name=} ({old_anki_parentdeck_name=} was removed from the stack)"
                        )
                # Check if the subdeck heading depth is at the expected depth
                if len(anki_decks_stack) != (heading_depth - initial_heading_depth):
                    expected_heading_depth = (
                        len(anki_decks_stack) + initial_heading_depth
                    )
                    raise AnkiSubdeckUnexpectedDepth(
                        f"The found {anki_subdeck.name=} heading has an unexpected {heading_depth=} "
                        f"({expected_heading_depth=})"
                    )
                # Add all tags from the parent deck to the current subdeck
                anki_subdeck.tags = anki_subdeck.tags.union(
                    anki_decks_stack[-1].get_used_global_tags()
                )
                anki_subdeck.md_document_order = anki_deck_order_index
                anki_deck_order_index += 1
                anki_decks_stack.append(anki_subdeck)
                log.debug(
                    f"<= [anki deck stack] Append {anki_subdeck=}",
                )
                parse_state = ParseStateMarkdownDocument.INSIDE_ANKI_DECK
            elif possible_anki_note is not None:
                # A note was detected
                anki_note, heading_depth = possible_anki_note
                log.debug(f">> Detected {anki_note=}")
                # Check if the note heading depth is at a supported depth
                if heading_depth > (len(anki_decks_stack) + 1) or heading_depth < (
                    initial_heading_depth + 1
                ):
                    raise AnkiNoteUnexpectedDepth(
                        f"The found {anki_note=} has an unexpected {heading_depth=} "
                        f"({initial_heading_depth + 1=}<=expected_depth<={len(anki_decks_stack) + 1=})"
                    )
                # In case the note belongs to another deck append finished subdecks to the final list
                parent_deck_removed = False
                while len(anki_decks_stack) > (heading_depth - initial_heading_depth):
                    anki_subdeck = anki_decks_stack.pop()
                    anki_decks.append(anki_subdeck)
                    log.debug(
                        f"<= [anki decks | new anki note at smaller heading depth] Append {anki_subdeck=}",
                    )
                    parent_deck_removed = True
                # If a parent deck was removed create a copy of the deck to preserve order
                if parent_deck_removed:
                    anki_subdeck = anki_decks_stack.pop()
                    anki_decks.append(anki_subdeck)
                    log.debug(
                        f"<= [anki decks | parent deck was removed] Append {anki_subdeck=}",
                    )
                    anki_deck = copy.deepcopy(anki_subdeck)
                    anki_deck.notes = []
                    anki_deck.md_document_order = anki_deck_order_index
                    anki_deck_order_index += 1
                    anki_decks_stack.append(anki_deck)
                    log.debug(f"<= [anki deck stack] Append {anki_deck=}")
                # Append the note to the current anki deck
                anki_deck = anki_decks_stack[-1]
                anki_deck.notes.append(anki_note)
                log.debug(
                    f"<= [anki deck | new anki note] Append {anki_note=} ({anki_deck.name=}, {anki_deck.guid=})",
                )
                parse_state = ParseStateMarkdownDocument.ANKI_NOTE_QUESTION
            elif parse_state == ParseStateMarkdownDocument.INSIDE_ANKI_DECK:
                # Update the current anki deck description
                anki_deck = anki_decks_stack[-1]
                anki_deck.description += empty_lines * "\n" + line
                empty_lines = 0
                log.debug(
                    f"=> Found/Update {anki_deck.description=} ({anki_deck.name=}, {anki_deck.guid=})",
                )
            elif (
                parse_state == ParseStateMarkdownDocument.ANKI_NOTE_QUESTION
                and line_stripped == MD2ANKI_MD_PP_ANKI_NOTE_QUESTION_ANSWER_SEPARATOR
            ):
                # Append read anki note answer to question and reset answer
                anki_note = anki_decks_stack[-1].notes[-1]
                anki_note.question = (
                    f"{anki_note.question.rstrip()}\n\n{anki_note.answer.lstrip()}"
                )
                anki_note.answer = ""
                log.debug(
                    f"=> Found question answer seperator/Update {anki_note.question=} ({anki_note.guid=})",
                )
                # Switch parse state to make sure future separators are being kept in the answer
                parse_state = ParseStateMarkdownDocument.ANKI_NOTE_ANSWER
            elif (
                parse_state == ParseStateMarkdownDocument.ANKI_NOTE_QUESTION
                or parse_state == ParseStateMarkdownDocument.ANKI_NOTE_ANSWER
            ):
                anki_note = anki_decks_stack[-1].notes[-1]
                anki_note.answer += empty_lines * "\n" + line
                empty_lines = 0
                log.debug(f"=> Found/Update {anki_note.answer=}")

    # Append yet not appended anki decks (in the correct order!)
    for anki_deck in anki_decks_stack:
        anki_decks.append(anki_deck)
        log.debug(f"<= [anki decks | reached eof] Append {anki_deck=}")

    if parse_state == ParseStateMarkdownDocument.LOOKING_FOR_ANKI_DECK_HEADING:
        raise NoAnkiDeckFoundException

    # Clean up descriptions of all anki decks as well as their notes
    for parent_anki_deck in anki_decks:
        parent_anki_deck.additional_file_dirs.extend(md_file_dirs)
        parent_anki_deck.description = parent_anki_deck.description.strip()
        for anki_note in parent_anki_deck.notes:
            anki_note.tags = anki_note.tags.union(
                parent_anki_deck.get_used_global_tags()
            )
            anki_note.question = anki_note.question.strip()
            anki_note.answer = anki_note.answer.strip()

    # Sort anki decks in order of their document order
    anki_decks.sort(key=lambda x: x.md_document_order)

    log.debug(
        f"=> Return anki decks: ({', '.join([f'{a.name} ({len(a.notes)})' for a in anki_decks])})",
    )

    return anki_decks



#!/usr/bin/env python3

# Open the entry point when program is executed using 'md2anki [...]' (check setup.py for more information)

# Local modules
from md2anki.entry_point import entry_point


def _main():
    entry_point()



#!/usr/bin/env python3

# Future import to support Python 3.9
from __future__ import annotations

# Internal packages
import logging
import re
from pathlib import Path
from re import Match
from typing import Callable, Final, Optional, Set, Tuple
from urllib.parse import urlparse, ParseResult

# Local modules
from md2anki.info.general import (
    MD2ANKI_MD_PP_MD2ANKI_TAG_PREFIX,
    MD2ANKI_MD_PP_MD2ANKI_TAG_SUFFIX,
)

# Logger
log = logging.getLogger(__name__)

# Constants
REGEX_MD_TAG: Final = re.compile(
    rf"`{MD2ANKI_MD_PP_MD2ANKI_TAG_PREFIX}(.*?){MD2ANKI_MD_PP_MD2ANKI_TAG_SUFFIX}`"
)
"""
Regex expression to parse a markdown tag notation: '`{=:tag list string:=}`'
The first group is the 'tag list string'.
"""

REGEX_MD_IMAGE_FILE: Final = re.compile(
    r"!\[(.*?)]\((.*?)\)(?:\{(?:\s*?width\s*?=(.+?)[\s,;]*?)?(?:\s*?height\s*?=(.+?)[\s,;]*?)?})?"
)
"""
Regex expression to parse a markdown image notation: '![alt text](source path){ width=100px, height=200px }'
The first group is the 'alt text' and the second one the 'source path' while optionally there is a third and fourth
group for the width and height if found.
"""

REGEX_CODE_BLOCK: Final = re.compile(
    r"(?:^[ \t]*[\n\r]|\A)(?P<indent>[ \t]*)```(?:\{(?P<language_pandoc>.+?)}|(?P<language>.+?))[\n\r]"
    r"(?P<code_with_indents>(?:(?:\1[^\r\n]*?|\s*?)[\r\n])+?)\1```"
    r"(?=(?:[ \t]*[\n\r]){2,}|(?:[ \t]*[\n\r])?[ \t]*\Z)",
    flags=re.MULTILINE,
)
"""
Regex expression to find code blocks: ```language\ncode\n```
The first group is a found indentation of the initial code block (no checks are being made!) or none, the second is the
language of the code in normal notation and is None if pandoc notation was found in the third group.
The fourth and last group is the actual code part (including indents!).
"""

REGEX_INLINE_CODE: Final = re.compile(r"(?<!\S)`([^`]+?)`(?:\{(.+?)})?(?!\S)")

REGEX_MATH_SECTION: Final = re.compile(
    r"\${2}((?:[^$]|\n)+?)\${2}|\$(?!\s)([^\n]+?)(?<!\s)(?<!\\)\$", flags=re.MULTILINE
)
"""
Regex expression to find math sections. The first part matches $$...$$ sections and the second part inline $...$
sections. To have the option to escape the $ symbol there is a negative lookbehind in front of the closing $ symbol
that checks if it was escaped and then moves on to the next $ symbol.
"""

# TODO Add update local files method
# TODO Add update (inline) code blocks method


def md_get_used_files(md_content: str) -> Set[Path | ParseResult]:
    """Get the used files of a Markdown text"""
    files: Final[Set[Path | ParseResult]] = set()

    def add_used_files(regex_group_match: Match) -> str:
        """Detect and add all image paths to the created set"""
        filepath = regex_group_match.group(2)
        possible_url = urlparse(filepath)
        if possible_url.scheme == "http" or possible_url.scheme == "https":
            files.add(possible_url)
        else:
            files.add(Path(filepath))
        return ""

    re.sub(REGEX_MD_IMAGE_FILE, add_used_files, md_content)
    return files


def md_update_local_filepaths(
    md_content: str, new_directory: Optional[Path] = None
) -> str:
    """Update all local filepaths to a custom directory"""

    def update_local_filepath(regex_group_match: Match):
        filepath = regex_group_match.group(2)
        # Ignore non local filepaths
        if filepath.startswith("https://") or filepath.startswith("http://"):
            return regex_group_match[0]
        file_name = Path(filepath).name
        return regex_group_match[0].replace(
            filepath,
            file_name
            if new_directory is None
            else str(new_directory.joinpath(file_name)),
        )

    return re.sub(REGEX_MD_IMAGE_FILE, update_local_filepath, md_content)


def md_update_images(
    md_content: str,
    image_replacer: Callable[[str, str, Optional[str], Optional[str]], str],
) -> str:
    def update_image(regex_group_match: Match) -> str:
        file_path = regex_group_match.group(2)
        file_description = regex_group_match.group(1)
        opt_image_width = regex_group_match.group(3)
        opt_image_height = regex_group_match.group(4)
        return image_replacer(
            file_path, file_description, opt_image_width, opt_image_height
        )

    return re.sub(REGEX_MD_IMAGE_FILE, update_image, md_content)


def md_update_code_parts(
    md_content: str,
    code_replacer: Callable[[str, bool, Optional[str], Optional[str]], str],
) -> str:
    """Update code parts `replacer(code, code_block, language, code_block_indent): updated code str`"""

    def code_block_replace(regex_group_match: Match):
        language_normal = regex_group_match.group(2)
        language_pandoc = regex_group_match.group(3)
        return code_replacer(
            regex_group_match.group(4),
            True,
            language_normal or language_pandoc,
            regex_group_match.group(1),
        )

    def inline_code_replace(regex_group_match: Match):
        return code_replacer(
            regex_group_match.group(1), False, regex_group_match.group(2), None
        )

    md_content = re.sub(REGEX_CODE_BLOCK, code_block_replace, md_content)
    md_content = re.sub(REGEX_INLINE_CODE, inline_code_replace, md_content)

    return md_content


def md_get_used_md2anki_tags(md_content: str) -> Set[str]:
    """Get the used (custom) md2anki tags of a Markdown text"""
    tags: Final[Set[str]] = set()

    def add_used_tags(regex_group_match: Match) -> str:
        """Detect and add all local found tags to the created set"""
        tag_strings = regex_group_match.group(1).split(",")
        for tag_string in tag_strings:
            tag = tag_string.strip()
            if " " in tag:
                old_tag = tag
                tag = tag.replace(" ", "_")
                log.warning(f"A tag with spaces {old_tag!r} was rewritten to {tag!r}")
            if len(tag) > 0:
                tags.add(tag)
        return ""

    re.sub(REGEX_MD_TAG, add_used_tags, md_content)
    return tags


def md_update_math_sections(md_content: str, replacer: Callable[[str, bool], str]):
    """Update math sections `replacer(math_section, block): updated str`"""

    def math_section_replace(regex_group_match: Match):
        if regex_group_match.group(1) is not None:
            return replacer(regex_group_match.group(1), True)
        else:
            return replacer(regex_group_match.group(2), False)

    md_content = re.sub(REGEX_MATH_SECTION, math_section_replace, md_content)
    return md_content


def md_update_generic_id_sections(
    md_content: str, section: Tuple[str, str], replacer: Callable[[str], str]
):
    """Update generic sections with a custom start and end `replacer(content): updated str`"""

    def generic_section_replace(regex_group_match: Match):
        return replacer(regex_group_match.group(1))

    regex_generic_section: Final = re.compile(rf"{section[0]}(.+?){section[1]}")
    md_content = re.sub(regex_generic_section, generic_section_replace, md_content)
    return md_content



#!/usr/bin/env python3

# Internal packages
import sys
import io
import unittest
from pathlib import Path
from typing import List, Tuple, Optional

# Append the module path for md2anki
sys.path.append(str(Path(__file__).parent.parent.joinpath("src")))

# Local modules
from md2anki.anki_deck import AnkiDeck
from md2anki.anki_note import AnkiNote
from md2anki.md_parser import (
    parse_possible_anki_deck_heading,
    parse_possible_anki_note_heading,
    parse_md_content_to_anki_deck_list,
    NoAnkiDeckFoundException,
    RootAnkiDeckWasSubdeckException,
    AnkiNoteUnexpectedDepth,
    AnkiSubdeckUnexpectedDepth,
)


class TestParsePossibleAnkiDeckHeading(unittest.TestCase):
    def setUp(self):
        self.md_content_list: List[str] = list()
        self.parent_deck_list: List[Optional[str]] = list()
        self.deck_id_list: List[bool] = list()
        self.results: List[Optional[Tuple[AnkiDeck, int, bool]]] = list()
        self.expected: List[Optional[Tuple[AnkiDeck, int, bool]]] = list()

        test_data: List[
            Tuple[str, Optional[str], bool, Optional[Tuple[AnkiDeck, int, bool]]]
        ] = [
            ("", None, False, None),
            ("just text", None, False, None),
            ("# Heading", None, False, (AnkiDeck(name="Heading"), 1, False)),
            (
                "# Heading (12345)",
                None,
                True,
                (AnkiDeck(name="Heading", guid=12345), 1, False),
            ),
            ("## Heading", None, False, (AnkiDeck(name="Heading"), 2, False)),
            (
                "## Subdeck: Heading",
                "parent deck",
                False,
                (AnkiDeck(name="parent deck::Heading"), 2, True),
            ),
            (
                "## Subdeck: Heading (12367)",
                "parent deck",
                True,
                (AnkiDeck(name="parent deck::Heading", guid=12367), 2, True),
            ),
            ("## Subdeck: Heading", None, False, (AnkiDeck(name=f"Heading"), 2, True)),
        ]

        for test_input, test_parent_deck, test_deck_id, test_expected in test_data:
            self.md_content_list.append(test_input)
            self.parent_deck_list.append(test_parent_deck)
            self.deck_id_list.append(test_deck_id)
            self.results.append(
                parse_possible_anki_deck_heading(
                    test_input, parent_deck_name=test_parent_deck
                )
            )
            self.expected.append(test_expected)

    def test_parsed_possible_anki_deck_same(self):
        for md_content, parent_deck, deck_id, result, expected in zip(
            self.md_content_list,
            self.parent_deck_list,
            self.deck_id_list,
            self.results,
            self.expected,
        ):
            with self.subTest(
                md_content=md_content, parent_deck_name=parent_deck, deck_id=deck_id
            ):
                if expected is None:
                    self.assertIsNone(
                        result,
                        f"Check if parsed anki deck None {result=}",
                    )
                elif result is None:
                    self.fail(f"{result=} was None but something else was {expected=}")
                else:
                    self.assertEqual(
                        result[0].name,
                        expected[0].name,
                        f"Check if parsed anki deck name {result[0].name=}=={expected[0].name=}",
                    )
                    if deck_id:
                        self.assertEqual(
                            result[0].guid,
                            expected[0].guid,
                            f"Check if parsed anki deck guid {result[0].guid=}=={expected[0].guid=}",
                        )
                    self.assertEqual(
                        result[1],
                        expected[1],
                        f"Check if parsed anki deck level {result[1]=}=={expected[1]=}",
                    )
                    self.assertEqual(
                        result[2],
                        expected[2],
                        f"Check if parsed anki deck subdeck {result[2]=}=={expected[2]=}",
                    )


class TestParsePossibleAnkiNoteQuestionHeader(unittest.TestCase):
    def setUp(self):
        self.md_content_list: List[str] = list()
        self.note_id_list: List[bool] = list()
        self.results: List[Optional[Tuple[AnkiNote, int]]] = list()
        self.expected: List[Optional[Tuple[AnkiNote, int]]] = list()

        test_data: List[Tuple[str, bool, Optional[Tuple[AnkiNote, int]]]] = [
            ("", False, None),
            ("just text", False, None),
            ("# Heading", False, None),
            ("## Heading", False, (AnkiNote(question="Heading"), 2)),
            (
                "## Heading (abcsdeef)",
                True,
                (AnkiNote(question="Heading", guid="abcsdeef"), 2),
            ),
        ]

        for test_input, test_note_id, test_expected in test_data:
            self.md_content_list.append(test_input)
            self.note_id_list.append(test_note_id)
            self.results.append(parse_possible_anki_note_heading(test_input))
            self.expected.append(test_expected)

    def test_parsed_possible_anki_note_question_same(self):
        for md_content, note_id, result, expected in zip(
            self.md_content_list, self.note_id_list, self.results, self.expected
        ):
            with self.subTest(md_content=md_content, note_id=note_id):
                if expected is None:
                    self.assertIsNone(
                        result,
                        f"Check if parsed anki note None ({result=})",
                    )
                elif result is None:
                    self.assertTrue(
                        False,
                        f"Result should never be None ({result=}) if expected is not None ({expected=})",
                    )
                else:
                    result_anki_note, result_anki_note_heading_depth = result
                    expected_anki_note, expected_anki_note_heading_depth = expected
                    self.assertEqual(
                        result_anki_note.question,
                        expected_anki_note.question,
                        f"Check if {result_anki_note.question=}=={expected_anki_note.question=}",
                    )
                    if note_id:
                        self.assertEqual(
                            result_anki_note.guid,
                            expected_anki_note.guid,
                            f"Check if {result_anki_note.guid=}=={expected_anki_note.guid=}",
                        )
                    self.assertEqual(
                        result_anki_note_heading_depth,
                        expected_anki_note_heading_depth,
                        f"Check if {result_anki_note_heading_depth=}=={expected_anki_note_heading_depth=}",
                    )


class TestParseMdContentToAnkiDeckList(unittest.TestCase):
    def setUp(self):
        self.md_content_list: List[str] = list()
        self.md_content_exception_list: List[str] = list()
        self.results: List[List[AnkiDeck]] = list()
        self.expected: List[List[AnkiDeck]] = list()
        self.expected_exception: List[Exception] = list()

        test_data: List[Tuple[str, List[AnkiDeck], Optional[Exception]]] = [
            (
                "",
                [],
                # type: ignore
                NoAnkiDeckFoundException,
            ),
            (
                "## Question",
                [],
                # type: ignore
                NoAnkiDeckFoundException,
            ),
            (
                "# Subdeck: Heading",
                [],
                # type: ignore
                RootAnkiDeckWasSubdeckException,
            ),
            (
                "# Heading\n\n### Question",
                [],
                # type: ignore
                AnkiNoteUnexpectedDepth,
            ),
            (
                "# Heading\n\n### Subdeck: Heading 2",
                [],
                # type: ignore
                AnkiSubdeckUnexpectedDepth,
            ),
            ("# Heading (1234)", [AnkiDeck(name="Heading", guid=1234)], None),
            (
                "# Heading (1234)\n\n" "## Question (abcdef)\n\nAnswer",
                [
                    AnkiDeck(
                        name="Heading",
                        guid=1234,
                        notes=[
                            AnkiNote(
                                question="Question", answer="Answer", guid="abcdef"
                            )
                        ],
                    )
                ],
                None,
            ),
            (
                "# Heading (12349)\n\n"
                "## Question 1 (abcdef)\n\nAnswer 1\n\n"
                "## Question 2 (abcdefg)\n\nMore question\n\n---\n\nAnswer 2",
                [
                    AnkiDeck(
                        name="Heading",
                        guid=12349,
                        notes=[
                            AnkiNote(
                                question="Question 1", answer="Answer 1", guid="abcdef"
                            ),
                            AnkiNote(
                                question="Question 2\n\nMore question",
                                answer="Answer 2",
                                guid="abcdefg",
                            ),
                        ],
                    )
                ],
                None,
            ),
            (
                "# Heading (1234)\n\n"
                "## Question 1 (abcdef)\n\nAnswer 1\n\n"
                "## Subdeck: Heading 2 (12345)\n\n"
                "### Question 2 (abcdefg)\n\nAnswer 2\n\n"
                "## Question 3 (abcdefgh)\n\nAnswer 3",
                [
                    AnkiDeck(
                        name="Heading",
                        guid=1234,
                        notes=[
                            AnkiNote(
                                question="Question 1", answer="Answer 1", guid="abcdef"
                            ),
                        ],
                    ),
                    AnkiDeck(
                        name="Heading::Heading 2",
                        guid=12345,
                        notes=[
                            AnkiNote(
                                question="Question 2", answer="Answer 2", guid="abcdefg"
                            ),
                        ],
                    ),
                    AnkiDeck(
                        name="Heading",
                        guid=1234,
                        notes=[
                            AnkiNote(
                                question="Question 3",
                                answer="Answer 3",
                                guid="abcdefgh",
                            ),
                        ],
                    ),
                ],
                None,
            ),
            (
                "# Heading (1234)\n\n`{=:global_tag:=}`\n\n"
                "## Question 1 (abcdef)\n\nAnswer 1\n\n"
                "## Subdeck: Heading 2 (12345)\n\n`{=:subdeck_tag:=}`\n\n"
                "### Question 2 (abcdefg)\n\nAnswer 2\n\n"
                "## Question 3 (abcdefgh)\n\nAnswer 3",
                [
                    AnkiDeck(
                        name="Heading",
                        guid=1234,
                        notes=[
                            AnkiNote(
                                question="Question 1",
                                answer="Answer 1",
                                guid="abcdef",
                                tags={"global_tag"},
                            ),
                        ],
                        tags={"global_tag"},
                    ),
                    AnkiDeck(
                        name="Heading::Heading 2",
                        guid=12345,
                        notes=[
                            AnkiNote(
                                question="Question 2",
                                answer="Answer 2",
                                guid="abcdefg",
                                tags={"global_tag", "subdeck_tag"},
                            ),
                        ],
                        tags={"global_tag", "subdeck_tag"},
                    ),
                    AnkiDeck(
                        name="Heading",
                        guid=1234,
                        notes=[
                            AnkiNote(
                                question="Question 3",
                                answer="Answer 3",
                                guid="abcdefgh",
                                tags={"global_tag"},
                            ),
                        ],
                        tags={"global_tag"},
                    ),
                ],
                None,
            ),
        ]

        for test_input, test_expected, test_expected_exception in test_data:
            if test_expected_exception is not None:
                self.md_content_exception_list.append(test_input)
                self.expected_exception.append(test_expected_exception)
            else:
                self.md_content_list.append(test_input)
                self.results.append(
                    parse_md_content_to_anki_deck_list(io.StringIO(test_input), [])
                )
                self.expected.append(test_expected)

    def test_parsed_anki_deck_list_same(self):
        for md_content, result, expected in zip(
            self.md_content_list, self.results, self.expected
        ):
            with self.subTest(md_content=md_content):
                for expected_anki_deck, result_anki_deck in zip(expected, result):
                    self.assertEqual(
                        expected_anki_deck.name,
                        result_anki_deck.name,
                        f"Check if {expected_anki_deck.name=}=={result_anki_deck.name=}",
                    )
                    self.assertEqual(
                        expected_anki_deck.guid,
                        result_anki_deck.guid,
                        f"Check if {expected_anki_deck.guid=}=={result_anki_deck.guid=}",
                    )
                    result_anki_deck_all_tags = (
                        result_anki_deck.get_used_global_tags().union(
                            result_anki_deck.tags
                        )
                    )
                    self.assertSetEqual(
                        expected_anki_deck.tags,
                        result_anki_deck_all_tags,
                        f"Check if {expected_anki_deck.tags=}=={result_anki_deck_all_tags=}",
                    )
                    for expected_anki_note, result_anki_note in zip(
                        expected_anki_deck.notes, result_anki_deck.notes
                    ):
                        self.assertEqual(
                            expected_anki_note.guid,
                            result_anki_note.guid,
                            f"Check if {expected_anki_note.guid=}=={result_anki_note.guid=}",
                        )
                        self.assertEqual(
                            expected_anki_note.question,
                            result_anki_note.question,
                            f"Check if {expected_anki_note.question=}=={result_anki_note.question=}",
                        )
                        self.assertEqual(
                            expected_anki_note.answer,
                            result_anki_note.answer,
                            f"Check if {expected_anki_note.answer=}=={result_anki_note.answer=}",
                        )
                        self.assertSetEqual(
                            expected_anki_note.tags,
                            result_anki_note.tags,
                            f"Check if {expected_anki_note.tags=}=={result_anki_note.tags=}",
                        )

    def test_parsed_anki_deck_list_exception_same(self):
        for md_content, expected_exception in zip(
            self.md_content_exception_list, self.expected_exception
        ):
            with self.subTest(md_content=md_content):
                # Ignore type error, is the correct type
                self.assertRaises(
                    # type: ignore
                    expected_exception,
                    parse_md_content_to_anki_deck_list,
                    io.StringIO(md_content),
                    [],
                )



#!/usr/bin/env python3

# Future import to support Python 3.9
from __future__ import annotations

# Internal modules
import sys
from pathlib import Path
from setuptools import setup, find_packages
from typing import Final

# Append the module path for md2anki standalone information files
sys.path.append(str(Path(__file__).parent.joinpath("src", "md2anki")))

# Local modules
# These imports will only import the self-contained constants of the files and not require the module dependencies
from info.general import (
    MD2ANKI_VERSION,
    MD2ANKI_NAME,
    MD2ANKI_AUTHOR,
    MD2ANKI_DESCRIPTION,
    MD2ANKI_URL,
    MD2ANKI_URL_BUG_TRACKER,
)
from info.files import (
    RELATIVE_RES_CSS_FILE_PATH,
    RELATIVE_RES_CSS_FILE_PATH_TYPE_ANSWER,
    RELATIVE_RES_CSS_FILE_PATH_TYPE_CLOZE,
)


with open("README.md", "r", encoding="utf-8") as f:
    long_description: Final = f.read()


setup(
    name=MD2ANKI_NAME,
    version=str(MD2ANKI_VERSION),
    author=MD2ANKI_AUTHOR,
    description=MD2ANKI_DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=MD2ANKI_URL,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.9",
    install_requires=[
        "beautifulsoup4>=4.12.2",
        "genanki>=0.13.0",
        "Markdown>=3.5",
        "Pygments>=2.16.1",
    ],
    package_data={
        MD2ANKI_NAME: [
            str(RELATIVE_RES_CSS_FILE_PATH),
            str(RELATIVE_RES_CSS_FILE_PATH_TYPE_ANSWER),
            str(RELATIVE_RES_CSS_FILE_PATH_TYPE_CLOZE),
        ],
    },
    entry_points={
        "console_scripts": [
            f"{MD2ANKI_NAME}={MD2ANKI_NAME}:_main",
        ],
    },
    project_urls={
        "Bug Tracker": MD2ANKI_URL_BUG_TRACKER,
    },
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python3

# Internal packages
import unittest
import sys
from pathlib import Path
from typing import List, Tuple

# Append the module path for md2anki
sys.path.append(str(Path(__file__).parent.parent.joinpath("src")))

# Local modules
from md2anki.html_util import fix_inline_code_p_tags


class TestFixInlineCodePTags(unittest.TestCase):
    def setUp(self):
        self.html_content_list: List[str] = list()
        self.results: List[str] = list()
        self.expected: List[str] = list()

        test_data: List[Tuple[str, str]] = [
            ("", ""),
            ("<p>Test</p>", "<p>Test</p>"),
            (
                '<div class="highlight_inline">Code</div>',
                '<div class="highlight_inline">Code</div>',
            ),
            (
                '<p>Test</p><div class="highlight_inline">Code</div><p>Test 2</p>',
                '<p class="highlight_inline_next">Test</p><div class="highlight_inline">Code</div><p>Test 2</p>',
            ),
        ]

        for test_input, test_expected in test_data:
            self.html_content_list.append(test_input)
            self.results.append(fix_inline_code_p_tags(test_input))
            self.expected.append(test_expected)

    def test_updated_html_same(self):
        for html_content, result, expected in zip(
            self.html_content_list, self.results, self.expected
        ):
            with self.subTest(html_content=html_content):
                self.assertEqual(
                    result,
                    expected,
                    f"Check if updated html {result=}=={expected=}",
                )



#!/usr/bin/env python3

# Installed packages
from bs4 import BeautifulSoup


def fix_inline_code_p_tags(html_content: str) -> str:
    """
    Find all p tags that are in front of inline code div tags and add a class to identify them using CSS rules.
    @param html_content: HTML content string
    @return: updated HTML content string
    """
    soup = BeautifulSoup(html_content, "html.parser")
    for inline_code_div in soup.findAll("div", {"class": "highlight_inline"}):
        for inline_code_div_prev in inline_code_div.previous_elements:
            if inline_code_div_prev.name == "p":
                inline_code_div_prev["class"] = inline_code_div_prev.get(
                    "class", []
                ) + ["highlight_inline_next"]
    return str(soup)



#!/usr/bin/env python3

# Internal packages
import hashlib
import json
import logging
import shutil
import tempfile
from pathlib import Path
from typing import List, Optional, Dict, Tuple

# Local modules
from md2anki.info.evaluate_code import (
    EVALUATE_CODE_MD2ANKI_PLACEHOLDER_CODE_STRING,
    EVALUATE_CODE_MD2ANKI_PREFIX_CODE_FILE_NAME,
)
from md2anki.info.general import MD2ANKI_NAME
from md2anki.subprocess import run_subprocess

# Logger
log = logging.getLogger(__name__)


class UnableToEvaluateCodeException(Exception):
    """Raised when not able to evaluate code"""

    pass


def evaluate_code_get_hash(
    code: str,
    program_binary: str,
    program_binary_args: List[str],
    additional_env: Optional[Dict[str, str]] = None,
    additional_hash: Optional[str] = None,
) -> str:
    sha256_hash_code_evaluation = hashlib.sha256()
    sha256_hash_code_evaluation.update(code.encode("utf-8"))
    sha256_hash_code_evaluation.update(program_binary.encode("utf-8"))
    for argument in program_binary_args:
        sha256_hash_code_evaluation.update(argument.encode("utf-8"))
    if additional_env is not None:
        sha256_hash_code_evaluation.update(
            json.dumps(additional_env, sort_keys=True).encode("utf-8")
        )
    if additional_hash is not None:
        sha256_hash_code_evaluation.update(additional_hash.encode("utf-8"))
    return sha256_hash_code_evaluation.hexdigest()


def evaluate_code_check_cache(
    cache_dir: Optional[Path],
    code: str,
    program_binary: str,
    program_binary_args: List[str],
    additional_env: Optional[Dict[str, str]] = None,
    additional_hash: Optional[str] = None,
) -> Tuple[bool, Optional[str]]:
    log.debug(f"Evaluate code check cache ({cache_dir=})")
    if cache_dir is None or not cache_dir.is_dir():
        log.debug("> False because cache dir None or not a dir")
        return False, None
    sha256_hash_code_evaluation = evaluate_code_get_hash(
        code,
        program_binary,
        program_binary_args,
        additional_env,
        additional_hash,
    )
    cached_code_evaluations_db = [x.name for x in cache_dir.glob("**/*") if x.is_dir()]
    if sha256_hash_code_evaluation in cached_code_evaluations_db:
        log.debug(f"> True because hit ({sha256_hash_code_evaluation=})")
        return True, sha256_hash_code_evaluation

    log.debug(f"> False because no hit ({sha256_hash_code_evaluation=})")
    return False, sha256_hash_code_evaluation


def evaluate_code_copy_and_get_cache(
    cache_dir: Path,
    target_dir: Path,
    cache_hash: str,
) -> str:
    log.debug(
        f"Evaluate code copy and get cache ({cache_dir=},{target_dir=},{cache_hash=})"
    )
    # Copy cached files
    cache_dir_code_evaluation = cache_dir.joinpath(cache_hash)
    if target_dir.is_dir():
        shutil.rmtree(target_dir)
    shutil.copytree(cache_dir_code_evaluation, target_dir)
    # Read command output
    cache_file_output_code_evaluation = cache_dir_code_evaluation.with_suffix(".txt")
    with open(cache_file_output_code_evaluation) as f:
        return f.read()


def evaluate_code_copy_and_store_cache(
    cache_dir: Optional[Path],
    target_dir: Path,
    command_output: str,
    cache_hash: Optional[str],
):
    log.debug(
        f"Evaluate code copy and store cache ({cache_dir=},{target_dir=},{cache_hash=})"
    )
    if cache_dir is None or cache_hash is None:
        # Do nothing if there is no cache directory
        return
    # Copy cached files
    cache_dir_code_evaluation = cache_dir.joinpath(cache_hash)
    if cache_dir_code_evaluation.is_dir():
        shutil.rmtree(cache_dir_code_evaluation)
    shutil.copytree(target_dir, cache_dir_code_evaluation)
    # Write command output
    cache_file_output_code_evaluation = cache_dir_code_evaluation.with_suffix(".txt")
    if cache_file_output_code_evaluation.is_file():
        cache_file_output_code_evaluation.unlink()
    with open(cache_file_output_code_evaluation, "w") as f:
        f.write(command_output)


def evaluate_code_in_subprocess(
    program: str,
    code: str,
    custom_program: Dict[str, List[str]],
    custom_program_args: Dict[str, List[List[str]]],
    dir_dynamic_files: Optional[Path] = None,
    keep_temp_files: bool = False,
    cache_dir: Optional[Path] = None,
    additional_env: Optional[Dict[str, str]] = None,
) -> Tuple[List[str], List[Path]]:
    """Return the command outputs and the found images."""
    log.debug(f"Evaluate code ({program=},{code=})")

    dir_path_temp = Path(tempfile.mkdtemp(prefix=f"{MD2ANKI_NAME}_evaluate_code_"))

    if program in custom_program:
        program_binaries = custom_program[program]
    else:
        raise UnableToEvaluateCodeException(f"Unsupported {program=} ({code=})")
    if program in custom_program_args:

        def insert_code_or_code_file(program_bin_arg: str) -> str:
            if program_bin_arg == EVALUATE_CODE_MD2ANKI_PLACEHOLDER_CODE_STRING:
                return code
            if program_bin_arg.startswith(EVALUATE_CODE_MD2ANKI_PREFIX_CODE_FILE_NAME):
                code_file_name = program_bin_arg[
                    len(EVALUATE_CODE_MD2ANKI_PREFIX_CODE_FILE_NAME) :
                ]
                if len(code_file_name) == 0:
                    raise RuntimeError(
                        f"Custom code file name had 0 length: {program_bin_arg!r}"
                    )
                with open(dir_path_temp.joinpath(code_file_name), "w") as temp_file:
                    temp_file.write(code)
                return code_file_name
            return program_bin_arg

        def update_list(program_bin_arg_dict_entry: List[str]) -> List[str]:
            return list(map(insert_code_or_code_file, program_bin_arg_dict_entry))

        program_binaries_args: List[List[str]] = list(
            map(update_list, custom_program_args[program])
        )
    else:
        program_binaries_args = list()
    while len(program_binaries_args) < len(program_binaries):
        program_binaries_args.append(list())

    try:
        results: List[str] = list()
        if len(program_binaries) != len(program_binaries_args):
            raise RuntimeError(
                f"Program binary ({len(program_binaries)}) and args list ({len(program_binaries_args)}) "
                "had different sizes"
            )
        previous_cache_hash: Optional[str] = None
        for program_binary, program_binary_args in zip(
            program_binaries, program_binaries_args
        ):
            # Check for cache
            cache_hit, cache_hash = evaluate_code_check_cache(
                cache_dir,
                code,
                program_binary,
                program_binary_args,
                additional_env,
                additional_hash=previous_cache_hash,
            )
            previous_cache_hash = cache_hash
            log.debug(f"Check cache {cache_dir=} ({cache_hit=},{cache_hash=})")
            if cache_hit and cache_hash is not None and cache_dir is not None:
                # Load cache
                p = evaluate_code_copy_and_get_cache(
                    cache_dir,
                    dir_path_temp,
                    cache_hash,
                )
                results.append(p)
            else:
                # Run subprocess
                try:
                    p = run_subprocess(
                        program_binary,
                        program_binary_args,
                        cwd=dir_path_temp,
                        additional_env=additional_env,
                    )
                except OSError as err:
                    log.error(
                        f"Error running the command {program_binary=} {program_binary_args=} cwd={dir_path_temp!r} {additional_env=} ({program=})"
                    )
                    raise err
                # Store cache
                evaluate_code_copy_and_store_cache(
                    cache_dir,
                    dir_path_temp,
                    p,
                    cache_hash,
                )
                results.append(p)
        images_list = list()
        if dir_dynamic_files is not None:
            for supported_media_files in [".svg", ".png", ".jpg"]:
                for glob_file in dir_path_temp.rglob(f"*{supported_media_files}"):
                    log.debug(
                        f"Copy created file {glob_file=} to {dir_dynamic_files=}",
                    )
                    shutil.copy2(glob_file, dir_dynamic_files)
                    images_list.append(dir_dynamic_files.joinpath(glob_file.name))
            log.debug(f"{images_list=} {results=}")
        return results, images_list

    finally:
        if not keep_temp_files:
            shutil.rmtree(dir_path_temp)
